module Sudoku(solve) where

import Data.List
import Data.Char

type Cell = Maybe Int
type Row a = [a]
type Grid a = [Row a]
type Options = [Cell]

boxes :: Grid a -> Grid a
boxes = deser . (map transpose) . ser
    where
    ser = (front 3) . map (front 3)
    deser = map concat . concat
    front _ [] = []
    front n xs = take n xs : front n (drop n xs)

filter_elem :: [Options] -> [Options]
filter_elem ops = [op `minus` single | op <- ops ]
    where
    single = concat (filter (\l -> length l == 1) ops)
    op1 `minus` op2 = if length op1 == 1 then op1 else op1\\op2

-- filter duplicate pairs
filter_pair :: [Options] -> [Options]
filter_pair x = [if (isInfixOf dup_pair xs) && (dup_pair /= xs) then xs\\dup_pair else xs | xs <- x]
    where
    dup_pair = concat $ getDups $ filter (\l -> length l == 2) x
    getDups t = t \\ nub t

-- pruneWith prune by (func) rule
prune :: Grid Options -> Grid Options
prune = (pruneWith filter_pair) . (pruneWith filter_elem)
    where
    pruneWith func = pruneBy id . pruneBy transpose . pruneBy boxes
        where
        pruneBy f = f . map func . f

allOptions :: Grid [a] -> [Grid a]
allOptions g = cardProd $ map cardProd g
    where 
    cardProd [] = [[]]
    cardProd (x:xs) = [y:ys | y <- x, ys <- cardProd xs]

-- start with the cell have fewest choices
try :: Grid Options -> [Grid Options]
try ops  = [rows_t ++ [row_t ++ [c] : row_aft] ++ rows_aft | c <- cs]
    where
    (rows_t, row:rows_aft) = break (any fit) ops
    (row_t, cs:row_aft) = break fit row
    fit t = (length t == len)
    len = minimum . filter(>1) . concat $ map (map length) ops

search :: Grid Options -> [Grid Cell]
search ops
    | any (any null) ops || not (success ops) = []
    | all (all (\l -> length l == 1)) ops = allOptions ops
    | otherwise = [s | ops' <- try ops, s <- search (arriveAt prune ops')]
    where
        arriveAt f x 
            | x == f x = x 
            | otherwise = arriveAt f (f x)
        success t = all good (t) && all good (transpose t) && all good (boxes t)
        good = noDup . concat . filter (\l -> length l == 1)
        noDup l = nub l == l

-- convert grid cell to sudoku string
deFormat :: Grid Cell -> String
deFormat cl = toStr $ concat cl
    where
    toStr [] = ""
    toStr (Nothing:xs) = "." ++ toStr xs
    toStr (Just a:xs) = [intToDigit a] ++ toStr xs

-- convert sudoku string to grid cell
format :: String -> Grid Cell
format xs 
    | length xs == 0 = [[]]
    | otherwise = (takeS 9 xs) : (format (drop 9 xs))
    where
        takeS :: Int -> String -> Row Cell
        takeS 0 _ = []
        takeS _ [] = []
        takeS _ [x] = convert x
        takeS n (x:xt) = convert x ++ takeS (n-1) xt
        convert x = if x == '.' then [Nothing] else [Just (digitToInt x)]

solve :: String -> [String]
solve str = mtxToStr matrix
    where
        matrix = search $ prune $ options $ format str
        mtxToStr mtx = [deFormat m | m <- mtx]
        options = map (map cur)
        cur v = if v == Nothing then [Just i | i <- [1..9]] else [v]